/*
 * Image codec. Author: Svjatoslav Agejenko, svjatoslav@svjatoslav.eu
 * This project is released under Creative Commons Zero (CC0) license.
 */
package eu.svjatoslav.imagesqueeze.codec;

import eu.svjatoslav.commons.data.BitInputStream;
import eu.svjatoslav.commons.data.BitOutputStream;

import java.awt.image.BufferedImage;
import java.io.*;

/**
 * Main class representing compressed image.
 */

public class Image {

    public ImageMetaData metaData;

    public BufferedImage bufferedImage;

    private ImageEncoder encoder;

    public Image() {
    }

    /**
     * Initialize imagesqueeze image based on {@link BufferedImage}.
     * {@link BufferedImage} must be of type BufferedImage.TYPE_3BYTE_BGR .
     */
    public Image(final BufferedImage image) {

        bufferedImage = image;
        metaData = new ImageMetaData();

        metaData.version = 1;
        metaData.width = image.getWidth();
        metaData.height = image.getHeight();
    }

    /**
     * Initialize empty imagesqueeze image.
     */
    public Image(final int width, final int height) {

        bufferedImage = new BufferedImage(width, height,
                BufferedImage.TYPE_3BYTE_BGR);
        metaData = new ImageMetaData();

        metaData.version = 1;
        metaData.width = width;
        metaData.height = height;
    }

    /**
     * Load ImgSqz image from {@link File}.
     */
    public void loadImage(final File source) throws IOException {

        final byte[] fileContent = new byte[(int) source.length()];

        try (FileInputStream fileInputStream = new FileInputStream(source)) {
            if (fileInputStream.read(fileContent) != fileContent.length)
                throw new RuntimeException("Failed to read file content");
        }

        final ByteArrayInputStream inputStream = new ByteArrayInputStream(
                fileContent);

        loadImage(inputStream);
    }

    /**
     * Load ImgSqz image from {@link InputStream}.
     */
    public void loadImage(final InputStream source) throws IOException {
        final BitInputStream bitInputStream = new BitInputStream(source);

        metaData = new ImageMetaData();
        metaData.load(bitInputStream);

        bufferedImage = new BufferedImage(metaData.width, metaData.height,
                BufferedImage.TYPE_3BYTE_BGR);

        final ImageDecoder imageDecoder = new ImageDecoder(this, bitInputStream);

        imageDecoder.decode();
    }

    /**
     * Save image into ImgSqz file format.
     */
    public void saveImage(final File file) throws IOException {
        final ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        saveImage(outputStream);

        final byte[] buffer = outputStream.toByteArray();
        try (FileOutputStream fileOutputStream = new FileOutputStream(file)) {
            fileOutputStream.write(buffer);
        }
    }

    /**
     * Save image into ImgSqz file format.
     */
    private void saveImage(final OutputStream outputStream) throws IOException {

        final BitOutputStream bitOutputStream = new BitOutputStream(
                outputStream);

        metaData.save(bitOutputStream);

        if (encoder == null) {
            encoder = new ImageEncoder(this);
        }

        encoder.encode(bitOutputStream);

        bitOutputStream.finishByte();
    }

}
